-- tolua: tag methods implemented in lua
-- This code is equivalent to the C tag method 
-- implementation used by tolua.
-- Written by Waldemar Celes
-- TeCGraf/PUC-Rio
-- Jul 1998
-- $Id: tm.lua,v 1.1 1998/11/24 16:10:28 celes Exp $

-- This code is free software; you can redistribute it and/or modify it.
-- The software provided hereunder is on an "as is" basis, and
-- the author has no obligation to provide maintenance, support, updates,
-- enhancements, or modifications. 

$debug

-- Inheritance
-- get table
function oo_gettable (table, base, index)
 while type(base)=='table' do
  local value = rawgettable(base,index)
  if value then
   return value
  elseif type(index) == 'number' and rawgettable(base,'operator_get') then
   return rawgettable(base,'operator_get')(table,index)
  else
   local get = rawgettable(base,".get")
   if get then
    local f = rawgettable(get,index)
    if f then
     return f(table)
    end
   end
  end
  base = rawgettable(base,".base")
 end
 return nil
end

-- set table
function oo_settable (table, base, index, value)
 while type(base)=='table' do
  local set = rawgettable(base,".set")
  if set then
   local f = rawgettable(set,index)
   if f then
    f(table,value)
    return 1
   end
  end
  base = rawgettable(base,".base")
 end
 return 0
end

-- Class tag methods
-- get table
function class_index (table, index)
 return oo_gettable(table,table,index)
end

-- set table
function class_settable (table, index, value)
 if oo_settable(table,table,index,value) == 0 then
  rawsettable(table,index,value)
 end
end

-- Instance tag methods
-- get table
function instance_gettable (table, index)
 local classtable = getglobal(".tolua_class")
 local base = classtable[tag(table)]
 local matetable = getglobal(".tolua_mate")
 if matetable[table] and matetable[table][index] then
  return matetable[table][index]
 else
  return oo_gettable(table,base,index)
 end
end

-- set table
function instance_settable (table, index, value)
 local classtable = getglobal(".tolua_class")
 local matetable = getglobal(".tolua_mate")
 local base = classtable[tag(table)]
 if type(index) == 'number' and rawgettable(base,'operator_set') then
  rawgettable(base,'operator_set')(table,index,value)
 elseif oo_settable(table,base,index,value) == 0 then
  local mate = matetable[table] -- should be 'void*'
  if not mate then
   mate = {}
   matetable[table] = mate
  end
  mate[index] = value
 end
end

-- gc
function instance_gc (obj)

end

-- operators
-- general operators
function instance_operator (op1, op2, event)
 local f = "operator_"..event
 return op1[f](op1,op2) 
end

-- relational operator
function instance_relational (op1, op2, event)
 local f = "operator_"..event
 local r = op1[f](op1,op2)
 if r==0 then
  return nil
 else
  return r
 end
end


-- Lua instance tag methods
-- index
function linstance_index (table, index)
 local classtable = getglobal(".tolua_class")
 local base = classtable[tag(table)]
 return oo_gettable(table,base,index)
end

-- Module tag methods
-- get table
function module_index (table, index)
 local get = rawgettable(table,".get")
 if get then
  local f = rawgettable(get,index)
  if f then
   return f()
  end
 end
 return nil
end
  
-- set table
function module_settable (table, index, value)
 local set = rawgettable(table,".set")
 if set then
  local f = rawgettable(set,index)
  if f then
   f(value)
   return
  end
 end
 rawsettable(table,index,value)
end

-- Global tag methods
-- get global
function global_getglobal (varname, value)
 local f = rawgettable(value,".get")
 return f()
end
 
-- set global
function global_setglobal (varname, value, newvalue)
 local f = rawgettable(value,".set")
 if not f then
  error("value of const variable cannot be changed")
 else
  f(newvalue)
 end
end

-- Indirect methods
-- getglobal
function indirect_getglobal (varname, value)
 return value[".module"][varname]
end
function indirect_setglobal (varname, value, newvalue)
 value[".module"][varname] = newvalue
end

